/******************************************************************************/
/*                                 M o r s e                                  */
/*                                                                            */
/* * The Morse.cpp library (ND-292012).                                       */
/* * Based on ITU-R M.1677-1 - International Morse code                       */
/*   Only the "single" characters in chapter 1.1.1, 1.1.2 and 1.1.3 are       */
/*   implemented, except the accented e.                                      */
/*                                                                            */
/* Revision log                                                               */
/* -------------------------------------------------------------------------- */
/* A00 - 2016.10.22 - JOO: Original version.                       1.0.0.   0 */
/* A01 - 2019.01.01 - JOO: Rewritten.                        (CVS) 1.1.0.   0 */
/*                                                                            */
/*                     Copyright (C) 2019 - Sintran Data                      */
/******************************************************************************/

#include "Morse.h"

// The table with all the legal morse characters and the corresponding morse code as a binary number.
// The table must be sorted on morseChar in ascending order, since we do a binary search in the table.
// The binary morse code is with 0 for a dot and 1 for a dash. The first 1 is not part of the morse code.
// The binary morse code can not be zeros only !
// Example: The letter C (-.-.) will have this binary morse code: B00011010
const morseElement Morse::morseTable[] =
{
		{'"' , B01010010}, // .-..-.
		{'\'', B01011110}, // .----.
		{'(' , B00110110}, // -.--.
		{')' , B01101101}, // -.--.-
		{'*' , B00011001}, // -..-
		{'+' , B00101010}, // .-.-.
		{',' , B01110011}, // --..--
		{'-' , B01100001}, // -....-
		{'.' , B01010101}, // .-.-.-
		{'/' , B00110010}, // -..-.
		{'0' , B00111111}, // -----
		{'1' , B00101111}, // .----
		{'2' , B00100111}, // ..---
		{'3' , B00100011}, // ...--
		{'4' , B00100001}, // ....-
		{'5' , B00100000}, // .....
		{'6' , B00110000}, // -....
		{'7' , B00111000}, // --...
		{'8' , B00111100}, // ---..
		{'9' , B00111110}, // ----.
		{':' , B01111000}, // ---...
		{'=' , B00110001}, // -...-
		{'?' , B01001100}, // ..--..
		{'@' , B01011010}, // .--.-.
		{'A' , B00000101}, // .-
		{'B' , B00011000}, // -...
		{'C' , B00011010}, // -.-.
		{'D' , B00001100}, // -..
		{'E' , B00000010}, // .
		{'F' , B00010010}, // ..-.
		{'G' , B00001110}, // --.
		{'H' , B00010000}, // ....
		{'I' , B00000100}, // ..
		{'J' , B00010111}, // .---
		{'K' , B00001101}, // -.-
		{'L' , B00010100}, // .-..
		{'M' , B00000111}, // --
		{'N' , B00000110}, // -.
		{'O' , B00001111}, // ---
		{'P' , B00010110}, // .--.
		{'Q' , B00011101}, // --.-
		{'R' , B00001010}, // .-.
		{'S' , B00001000}, // ...
		{'T' , B00000011}, // -
		{'U' , B00001001}, // ..-
		{'V' , B00010001}, // ...-
		{'W' , B00001011}, // .--
		{'X' , B00011001}, // -..-
		{'Y' , B00011011}, // -.--
		{'Z' , B00011100}  // --..
};

// Public functions ==========================================================//
/*============================================================================*/
/* * Constructor for the Morse object.                                        */
/* * The digital output pin will be set to the default (M_DEFAULT_PIN) value. */
/*   \param[in] duration  - The dot duration in ms.                           */
/*   \param[in] frequency - The tone to be used in Hz.                        */
/*   \return none.                                                            */
/*============================================================================*/
Morse::Morse(int duration, int frequency)
{
	_pin       = M_DEFAULT_PIN;
	_duration  = duration;
	_frequency = frequency;
	return;
}

/*============================================================================*/
/* * Setup the digital output pin, ready for usage (if defined).              */
/* * If M_DEBUG is defined: Print configuration and validate the morseTable.  */
/*   \param[in] pin - The digital output pin to use.                          */
/*   \return none.                                                            */
/*============================================================================*/
void Morse::begin(byte pin)
{
	_pin = pin;

	if (_pin != M_DEFAULT_PIN)
		pinMode(_pin, OUTPUT);

#ifdef M_DEBUG
	{
		int  i;

		trace("pin=%d, duration=%d, frequency=%d", _pin, _duration, _frequency);

		for (i = 0; i < (int)(sizeof(morseTable) / sizeof(morseElement) - 1); i++)
		{
			trace("%2d: Char: 0x%02X (%c), Code: 0x%02X - %s", i, morseTable[i].morseChar, morseTable[i].morseChar, morseTable[i].morseCode,
					(morseTable[i].morseChar >= morseTable[i + 1].morseChar) || (morseTable[i].morseCode == 0x00) ? "Error" : "OK");
		}
		trace("%2d: Char: 0x%02X (%c), Code: 0x%02X - %s", i, morseTable[i].morseChar, morseTable[i].morseChar, morseTable[i].morseCode,
				(morseTable[i].morseCode == 0x00) ? "Error" : "OK");
	}
#endif
	return;
}

/*============================================================================*/
/* * Setup the digital output pin, ready for usage (if defined).              */
/* * Changing the dot duration.                                               */
/*   \param[in] pin      - The digital output pin to use.                     */
/*   \param[in] duration - The dot duration in ms.                            */
/*   \return none.                                                            */
/*============================================================================*/
void Morse::begin(byte pin, int duration)
{
	_duration = duration;
	begin(pin);
	return;
}

/*============================================================================*/
/* * Setup the digital output pin, ready for usage (if defined).              */
/* * Changing both the dot duration and the tone frequency.                   */
/*   \param[in] pin       - The digital output pin to use.                    */
/*   \param[in] duration  - The dot duration in ms.                           */
/*   \param[in] frequency - The tone to be used in Hz.                        */
/*   \return none.                                                            */
/*============================================================================*/
void Morse::begin(byte pin, int duration, int frequency)
{
	_duration  = duration;
	_frequency = frequency;
	begin(pin);
	return;
}

/*============================================================================*/
/* * Generate a morse dash with the correct duration and spacing.             */
/*   \param[in] none.                                                         */
/*   \return none.                                                            */
/*============================================================================*/
void Morse::dash()
{
#ifdef M_DEBUG
	Serial.print("-");
#endif

	if (_pin != M_DEFAULT_PIN)
	{
		tone(_pin, _frequency);
		wait(M_DURATION_DASH);
		noTone(_pin);
		wait(M_DURATION_SIGNAL);
	}
	return;
}

/*============================================================================*/
/* * Generate a morse dot with the correct duration and spacing.              */
/*   \param[in] none.                                                         */
/*   \return none.                                                            */
/*============================================================================*/
void Morse::dot(void)
{
#ifdef M_DEBUG
	Serial.print(".");
#endif

	if (_pin != M_DEFAULT_PIN)
	{
		tone(_pin, _frequency);
		wait(M_DURATION_DOT);
		noTone(_pin);
		wait(M_DURATION_SIGNAL);
	}
	return;
}

/*============================================================================*/
/* * Return the current (dot) duration in ms.                                 */
/*   \param[in] none.                                                         */
/*   \return Duration.                                                        */
/*============================================================================*/
int Morse::getDuration(void)
{
	return _duration;
}

/*============================================================================*/
/* * Return the current frequency in Hz.                                      */
/*   \param[in] none.                                                         */
/*   \return Frequency.                                                       */
/*============================================================================*/
int Morse::getFrequency(void)
{
	return _frequency;
}

/*============================================================================*/
/* * Return the current output pin.                                           */
/*   \param[in] none.                                                         */
/*   \return Output pin.                                                      */
/*============================================================================*/
byte Morse::getPin(void)
{
	return _pin;
}

/*============================================================================*/
/* * Convert text into morse code ('.' and '-') and return it as a string.    */
/*   \param[out] str - Pointer to the output string buffer.                   */
/*   \param[in ] max - The max size (in bytes) for the string buffer.         */
/*   \param[in ] txt - The text that will be converted into morse code.       */
/*   \return Pointer to the morse code string (str).                          */
/*============================================================================*/
char* Morse::string(char* str, int max, const char* txt)
{
	byte bitNo;	// Current bit in the morse code
	byte code;	// Current morse code
	int  sindx;	// Current index into the output morse string
	int  tindx;	// Current index into the source text string

	max--;		// Leave space for the terminating zero in the output string
	sindx = 0;
	for (tindx = 0; tindx < (int)strlen(txt); tindx++)
	{
		if (sindx >= max)
			break;

		if (txt[tindx] == ' ')
			putChar(str, &sindx, max, ' ');
		else
		{
			if((code = getMorseCode(txt[tindx])) != 0)
			{
				for (bitNo = findStartOfMorseCode(code); bitNo > 0; )
				{
					if (bitRead(code, --bitNo) == 0)
						putChar(str, &sindx, max, '.');
					else
						putChar(str, &sindx, max, '-');
				}
				putChar(str, &sindx, max, ' ');
			}
		}
	}
	str[sindx] = 0x00;
	return str;
}

/*============================================================================*/
/* * Make a delay according to the given delay type.                          */
/*   \param[in] type - Delay type M_DURATION_DOT/DASH/SIGNAL/LETTER/WORD.     */
/*   \return none.                                                            */
/*============================================================================*/
void Morse::wait(int type)
{
	if (type > 0)
		delay(_duration * type);
	return;
}

/*============================================================================*/
/* * Overrides the low level write method in the Print class.                 */
/* * This method "prints" one morse character on the defined output pin.      */
/*   \param[in] ch - Any printable character.                                 */
/*   \return One if a morse character is "printed" else zero.                 */
/*============================================================================*/
size_t Morse::write(uint8_t ch)
{
	byte bitNo;	// Current bit in the morse code
	byte code;	// Current morse code

	if (ch == ' ')
	{
		wait(M_DURATION_WORD);
		return 1;
	}

	if((code = getMorseCode(ch)) == 0)
		return 0;

	for (bitNo = findStartOfMorseCode(code); bitNo > 0; )
	{
		if (bitRead(code, --bitNo) == 0)
			dot();
		else
			dash();
	}

#ifdef M_DEBUG
	Serial.println();
#endif
	wait(M_DURATION_LETTER - 1); // One dot spacing is already done by dot or dash functions
	return 1;
}

// Private functions =========================================================//
/*============================================================================*/
/* * Scan for the first '1' which marks the beginning of the morse code.      */
/*   \param[in] code - The binary morse code.                                 */
/*   \return Bit number for the first '1'.                                    */
/*============================================================================*/
byte Morse::findStartOfMorseCode(byte code)
{
	byte bitNo;

	for (bitNo = 7; (bitNo > 0) && (bitRead(code, bitNo) != 1); bitNo--) {}
	return bitNo;
}

/*============================================================================*/
/* * Do a binary search for the character and return the binary morse code.   */
/*   \param[in] ch - Any printable character.                                 */
/*   \return The binary morse code, zero if not found.                        */
/*============================================================================*/
byte Morse::getMorseCode(char ch)
{
	byte F; // The first  element in the current search region
	byte L; // The last   element in the current search region
	byte M; // The middle element in the current search region

	if (!isPrintable(ch)) // Ignore if the character is not printable
		return 0;

	if (isLowerCase(ch)) // Convert to uppercase, if required
		ch -= ('a' - 'A');

	// Start with a search region that is the entire morseTable
	F = 0;
	L = (sizeof(morseTable) / sizeof(morseElement)) - 1;

	while (true)
	{
		M = ((L - F) / 2) + F; // Find the element in the middle of the search region
#ifdef M_DEBUG
		trace("ch=%c, F=%2d, M=%2d, L=%2d, [M]=%c, 0x%02X", ch, F, M, L, morseTable[M].morseChar, morseTable[M].morseCode);
#endif
		if (ch == morseTable[M].morseChar) // Yes, spot on, return the binary morse code
			return morseTable[M].morseCode;

		if (F >= L) // If the search region limits meets, we did not find the wanted character
			break;

		if (ch < morseTable[M].morseChar) // Adjust the search region limit(s)
			L = M > 0 ? M - 1 : M;
		else
			F = M + 1;
	}
	return 0;
}

/*============================================================================*/
/* * Insert a character into the string at given index, increment the index.  */
/* * Make sure we are not outside the given string buffer.                    */
/*   \param[   out] str   - Pointer to the string buffer.                     */
/*   \param[in/out] sindx - Current index into the string buffer              */
/*   \param[in    ] max   - The max size (in bytes) for the string buffer.    */
/*   \return none.                                                            */
/*============================================================================*/
void Morse::putChar(char* str, int* sindx, int max, char ch)
{
	if (*sindx < max)
		str[(*sindx)++] = ch;
}

#ifdef M_DEBUG
/*============================================================================*/
/* * Print a trace line on the standard serial port, only used for debugging. */
/*   \param[in] format - The trace line to write.                             */
/*   \param[in] ...    - Parameter(s) required by the format print line.      */
/*   \return none.                                                            */
/*============================================================================*/
void Morse::trace(const char* format, ...)
{
	va_list args;
	char str[80];

	va_start(args, format);
	vsnprintf (str, sizeof(str), format, args);
	va_end(args);
	Serial.println(str);
	return;
}
#endif
